# Generated by Django 2.1.4 on 2019-01-03 10:44

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import mptt.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Callback',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, null=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, null=True, verbose_name='Date Updated')),
                ('hash', models.CharField(max_length=200, unique=True, verbose_name='Hash')),
                ('method', models.CharField(max_length=200, verbose_name='Callback Method')),
                ('hooking_cls', models.CharField(max_length=200, verbose_name='HookingClass')),
                ('enabled', models.BooleanField(default=True, verbose_name='Enabled')),
            ],
            options={
                'verbose_name': 'Callback',
                'verbose_name_plural': 'Callbacks',
            },
        ),
        migrations.CreateModel(
            name='State',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, null=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, null=True, verbose_name='Date Updated')),
                ('slug', models.SlugField(blank=True, null=True, unique=True)),
                ('label', models.CharField(max_length=50)),
                ('description', models.CharField(blank=True, max_length=200, null=True, verbose_name='Description')),
            ],
            options={
                'verbose_name': 'State',
                'verbose_name_plural': 'States',
            },
        ),
        migrations.CreateModel(
            name='TransitionApproval',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, null=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, null=True, verbose_name='Date Updated')),
                ('field_name', models.CharField(max_length=200, verbose_name='Field Name')),
                ('object_id', models.CharField(max_length=50, verbose_name='Related Object')),
                ('transaction_date', models.DateTimeField(blank=True, null=True)),
                ('status', models.IntegerField(choices=[(0, 'Pending'), (1, 'Approved'), (2, 'Rejected')], default=0, verbose_name='Status')),
                ('skip', models.BooleanField(default=False, verbose_name='Skip')),
                ('priority', models.IntegerField(default=0, verbose_name='Priority')),
                ('enabled', models.BooleanField(default=True, verbose_name='Enabled?')),
                ('cloned', models.BooleanField(default=False, verbose_name='Cloned?')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType', verbose_name='Content Type')),
                ('destination_state', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='transition_approvals_as_destination', to='river.State', verbose_name='Next State')),
                ('groups', models.ManyToManyField(to='auth.Group', verbose_name='Groups')),
            ],
            options={
                'verbose_name': 'Transition Approval',
                'verbose_name_plural': 'Transition Approvals',
            },
        ),
        migrations.CreateModel(
            name='TransitionApprovalMeta',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, null=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, null=True, verbose_name='Date Updated')),
                ('field_name', models.CharField(max_length=200, verbose_name='Field Name')),
                ('priority', models.IntegerField(default=0, null=True, verbose_name='Priority')),
                ('action_text', models.TextField(blank=True, max_length=200, null=True, verbose_name='Action Text')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType', verbose_name='Content Type')),
                ('destination_state', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='transition_approval_meta_as_destination', to='river.State', verbose_name='Next State')),
                ('groups', models.ManyToManyField(blank=True, to='auth.Group', verbose_name='Groups')),
                ('parents', models.ManyToManyField(blank=True, db_index=True, related_name='children', to='river.TransitionApprovalMeta', verbose_name='parents')),
                ('permissions', models.ManyToManyField(blank=True, to='auth.Permission', verbose_name='Permissions')),
                ('source_state', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='transition_approval_meta_as_source', to='river.State', verbose_name='Source State')),
            ],
            options={
                'verbose_name': 'Transition Approval Meta',
                'verbose_name_plural': 'Transition Approval Meta',
            },
        ),
        migrations.AddField(
            model_name='transitionapproval',
            name='meta',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='proceedings', to='river.TransitionApprovalMeta', verbose_name='Meta'),
        ),
        migrations.AddField(
            model_name='transitionapproval',
            name='permissions',
            field=models.ManyToManyField(to='auth.Permission', verbose_name='Permissions'),
        ),
        migrations.AddField(
            model_name='transitionapproval',
            name='previous',
            field=mptt.fields.TreeOneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='next_transition', to='river.TransitionApproval', verbose_name='Previous Transition'),
        ),
        migrations.AddField(
            model_name='transitionapproval',
            name='source_state',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='transition_approvals_as_source', to='river.State', verbose_name='Source State'),
        ),
        migrations.AddField(
            model_name='transitionapproval',
            name='transactioner',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='Transactioner'),
        ),
        migrations.AlterUniqueTogether(
            name='transitionapprovalmeta',
            unique_together={('content_type', 'field_name', 'source_state', 'destination_state', 'priority')},
        ),
    ]



import inspect

from river.core.classworkflowobject import ClassWorkflowObject
from river.core.instanceworkflowobject import InstanceWorkflowObject
from river.core.workflowregistry import workflow_registry


class RiverObject(object):

    def __init__(self, owner, field_name):
        self.owner = owner
        self.field_name = field_name
        self.is_class = inspect.isclass(owner)

    def __getattr__(self, field_name):
        cls = self.owner if self.is_class else self.owner.__class__
        if field_name not in workflow_registry.workflows[id(cls)]:
            raise Exception("Workflow with name:%s doesn't exist for class:%s" % (field_name, cls.__name__))
        if self.is_class:
            return ClassWorkflowObject(self.owner, field_name, self.field_name)
        else:
            return InstanceWorkflowObject(self.owner, field_name, self.field_name)

    def all(self, cls):
        return list([getattr(self, field_name) for field_name in workflow_registry.workflows[id(cls)]])


import logging
from uuid import uuid4

from django.db.models import CASCADE
from django.db.models.signals import post_save

from river.core.riverobject import RiverObject
from river.core.workflowregistry import workflow_registry

try:
    from django.contrib.contenttypes.fields import GenericRelation
except ImportError:
    from django.contrib.contenttypes.generic import GenericRelation

from river.models.state import State
from river.models.transitionapproval import TransitionApproval

__author__ = 'ahmetdal'

from django.db import models

LOGGER = logging.getLogger(__name__)


class classproperty(object):
    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(instance) if instance else self.getter(owner)


class StateField(models.ForeignKey):
    def __init__(self, *args, **kwargs):
        self.field_name = None
        kwargs['null'] = True
        kwargs['blank'] = True
        kwargs['to'] = '%s.%s' % (State._meta.app_label, State._meta.object_name)
        kwargs['on_delete'] = kwargs.get('on_delete', CASCADE)
        kwargs['related_name'] = "rn" + str(uuid4()).replace("-", "")
        super(StateField, self).__init__(*args, **kwargs)

    def contribute_to_class(self, cls, name):
        @classproperty
        def river(_self):
            return RiverObject(_self, name)

        self.field_name = name

        self._add_to_class(cls, self.field_name + "_transitions", GenericRelation('%s.%s' % (TransitionApproval._meta.app_label, TransitionApproval._meta.object_name)))

        if id(cls) not in workflow_registry.workflows:
            self._add_to_class(cls, "river", river)

        super(StateField, self).contribute_to_class(cls, name)

        if id(cls) not in workflow_registry.workflows:
            post_save.connect(_post_save, self.model, False, dispatch_uid='%s_%s_riverstatefield_post' % (self.model, name))

        workflow_registry.add(self.field_name, cls)


    @staticmethod
    def _add_to_class(cls, key, value, ignore_exists=False):
        if ignore_exists or not hasattr(cls, key):
            cls.add_to_class(key, value)


def _post_save(sender, instance, created, *args, **kwargs):  # signal, sender, instance):
    for workflow in instance.river.all(instance.__class__):
        if created:
            workflow.initialize_approvals()
        if not workflow.get_state():
            init_state = getattr(instance.__class__.river, workflow.name).initial_state
            workflow.set_state(init_state)
            instance.save()



from river.config import app_config
from river.hooking.backends.loader import load_callback_backend
from river.hooking.transition import PostTransitionHooking
from river.models.callback import Callback
from river.tests.base_test import BaseTestCase
from river.tests.models.factories import TestModelObjectFactory

__author__ = 'ahmetdal'


def test_callback(*args, **kwargs):
    pass


class DatabaseHookingBackendTest(BaseTestCase):
    def setUp(self):
        super(DatabaseHookingBackendTest, self).setUp()
        self.field_name = "my_field"
        self.initialize_standard_scenario()
        app_config.HOOKING_BACKEND_CLASS = 'river.hooking.backends.database.DatabaseHookingBackend'
        self.handler_backend = load_callback_backend()
        self.handler_backend.callbacks = {}

    def test_register(self):
        objects = TestModelObjectFactory.create_batch(2)

        self.assertEqual(0, Callback.objects.count())
        self.assertFalse('%s.%s_object%sfield%s' % (
            PostTransitionHooking.__module__, PostTransitionHooking.__name__, objects[1].pk,
            'my_field') in self.handler_backend.callbacks)

        self.handler_backend.register(PostTransitionHooking, test_callback, objects[1], self.field_name)

        self.assertEqual(1, Callback.objects.filter(hash='%s.%s_object%s_field_name%s' % (
            PostTransitionHooking.__module__, PostTransitionHooking.__name__, objects[1].pk, self.field_name)).count())
        self.assertTrue('%s.%s_object%s_field_name%s' % (
            PostTransitionHooking.__module__, PostTransitionHooking.__name__, objects[1].pk, self.field_name) in self.handler_backend.callbacks)
        self.assertEqual(test_callback.__name__, list(self.handler_backend.callbacks.values())[0].__name__)

    def test_register_in_multiprocessing(self):
        objects = TestModelObjectFactory.create_batch(2)

        from multiprocessing import Process, Queue

        self.handler_backend.register(PostTransitionHooking, test_callback, objects[1], self.field_name)
        self.assertEqual(1, Callback.objects.count())

        def worker2(q):
            second_handler_backend = load_callback_backend()
            handlers = second_handler_backend.get_callbacks(PostTransitionHooking, objects[1], self.field_name)
            q.put([f.__name__ for f in handlers])

        q = Queue()
        p2 = Process(target=worker2, args=(q,))

        p2.start()

        handlers = q.get(timeout=1)
        self.assertEqual(1, len(handlers))
        self.assertEqual(test_callback.__name__, handlers[0])

    def test_get_callbacks(self):
        objects = TestModelObjectFactory.create_batch(2)

        self.handler_backend.register(PostTransitionHooking, test_callback, objects[1], self.field_name)
        handlers = self.handler_backend.get_callbacks(PostTransitionHooking, objects[1], self.field_name)
        self.assertEqual(1, len(handlers))

        self.assertEqual(test_callback.__name__, handlers[0].__name__)

    def test_get_handlers_in_multiprocessing(self):
        objects = TestModelObjectFactory.create_batch(2)

        from multiprocessing import Process, Queue

        Callback.objects.update_or_create(
            hash='%s.%s_object%s_field_name%s' % (PostTransitionHooking.__module__, PostTransitionHooking.__name__, objects[1].pk, self.field_name),
            defaults={
                'method': '%s.%s' % (test_callback.__module__, test_callback.__name__),
                'hooking_cls': '%s.%s' % (PostTransitionHooking.__module__, PostTransitionHooking.__name__),
            }
        )

        def worker2(q):
            handlers = self.handler_backend.get_callbacks(PostTransitionHooking, objects[1], self.field_name)
            q.put([f.__name__ for f in handlers])

        q = Queue()
        p2 = Process(target=worker2, args=(q,))

        p2.start()

        handlers = q.get(timeout=1)
        self.assertEqual(1, len(handlers))
        self.assertEqual(test_callback.__name__, handlers[0])



import logging
from abc import abstractmethod

from river.hooking.backends.loader import callback_backend

__author__ = 'ahmetdal'

LOGGER = logging.getLogger(__name__)


class Hooking(object):

    @staticmethod
    def get_result_exclusions():
        return []

    @classmethod
    def dispatch(cls, workflow_object, field_name, *args, **kwargs):
        LOGGER.debug("Hooking %s is dispatched for workflow object %s and field name %s" % (cls.__name__, workflow_object, field_name))
        kwargs.pop('signal', None)
        kwargs.pop('sender', None)

        for callback in callback_backend.get_callbacks(cls, workflow_object, field_name, *args, **kwargs):
            exclusions = cls.get_result_exclusions()
            callback(workflow_object, field_name, *args, **{k: v for k, v in kwargs.items() if k not in exclusions})
            LOGGER.debug(
                "Hooking %s for workflow object %s and for field %s is found as method %s with args %s and kwargs %s" % (cls.__name__, workflow_object, field_name, callback.__name__, args, kwargs))

    @classmethod
    def register(cls, callback, workflow_object, field_name, override=False, *args, **kwargs):
        callback_backend.register(cls, callback, workflow_object, field_name, override, *args, **kwargs)

    @classmethod
    def get_hash(cls, workflow_object, field_name, *args, **kwargs):
        return 'object' + (str(workflow_object.pk) if workflow_object else '') + '_field_name' + field_name



from django.contrib.contenttypes.models import ContentType

from river.models.factories import StateObjectFactory, TransitionApprovalMetaFactory
from river.core.riverobject import RiverObject
from river.core.instanceworkflowobject import InstanceWorkflowObject
from river.core.classworkflowobject import ClassWorkflowObject
from river.tests.base_test import BaseTestCase
from river.tests.models.testmodel import TestModel

__author__ = 'ahmetdal'


class StateFieldTest(BaseTestCase):

    def test_injections(self):
        self.assertTrue(hasattr(TestModel, 'river'))
        self.assertIsInstance(TestModel.river, RiverObject)
        self.assertTrue(hasattr(TestModel.river, "my_field"))
        self.assertIsInstance(TestModel.river.my_field, ClassWorkflowObject)

        content_type = ContentType.objects.get_for_model(TestModel)

        state1 = StateObjectFactory.create(label="state1")
        state2 = StateObjectFactory.create(label="state2")

        TransitionApprovalMetaFactory.create(
            content_type=content_type,
            field_name="my_field",
            source_state=state1,
            destination_state=state2,
            priority=0
        )
        test_model = TestModel.objects.create()
        self.assertTrue(hasattr(test_model, "river"))
        self.assertIsInstance(test_model.river, RiverObject)
        self.assertTrue(hasattr(test_model.river, "my_field"))
        self.assertIsInstance(test_model.river.my_field, InstanceWorkflowObject)

        self.assertTrue(hasattr(test_model.river.my_field, "approve"))
        self.assertTrue(callable(test_model.river.my_field.approve))

        self.assertTrue(test_model.river.my_field.on_initial_state)
        self.assertFalse(test_model.river.my_field.on_final_state)

        self.assertEqual(state1, TestModel.river.my_field.initial_state)
        self.assertEqual(1, TestModel.river.my_field.final_states.count())
        self.assertEqual(state2, TestModel.river.my_field.final_states[0])



from django.db import models

from river.config import app_config

__author__ = 'ahmetdal'


class TransitionApprovalManager(models.Manager):
    def __init__(self, *args, **kwargs):
        super(TransitionApprovalManager, self).__init__(*args, **kwargs)

    def filter(self, *args, **kwarg):
        object = kwarg.pop('workflow_object', None)
        if object:
            kwarg['content_type'] = app_config.CONTENT_TYPE_CLASS.objects.get_for_model(object)
            kwarg['object_id'] = object.pk

        return super(TransitionApprovalManager, self).filter(*args, **kwarg)

    def update_or_create(self, *args, **kwarg):
        object = kwarg.pop('workflow_object', None)
        if object:
            kwarg['content_type'] = app_config.CONTENT_TYPE_CLASS.objects.get_for_model(object)
            kwarg['object_id'] = object.pk

        return super(TransitionApprovalManager, self).update_or_create(*args, **kwarg)



from datetime import datetime, timedelta

from river.models import TransitionApproval, APPROVED, TransitionApprovalMeta
from river.tests.base_test import BaseTestCase
from river.tests.models.factories import TestModelObjectFactory
from river.tests.models.testmodel import TestModel
from river.utils.error_code import ErrorCode
from river.utils.exceptions import RiverException

__author__ = 'ahmetdal'


class RiverTest(BaseTestCase):

    def test_get_on_approval_objects(self):
        self.initialize_standard_scenario()
        objects = TestModelObjectFactory.create_batch(2)

        on_approval_objects = TestModel.river.my_field.get_on_approval_objects(as_user=self.user1)
        self.assertEqual(2, on_approval_objects.count())
        self.assertEqual(objects[0], on_approval_objects[0])

        on_approval_objects = TestModel.river.my_field.get_on_approval_objects(as_user=self.user2)
        self.assertEqual(0, on_approval_objects.count())

        on_approval_objects = TestModel.river.my_field.get_on_approval_objects(as_user=self.user3)
        self.assertEqual(0, on_approval_objects.count())

        on_approval_objects = TestModel.river.my_field.get_on_approval_objects(as_user=self.user4)
        self.assertEqual(0, on_approval_objects.count())

    def test_get_available_states(self):
        self.initialize_standard_scenario()
        object = TestModelObjectFactory.create_batch(1)[0]
        available_states = object.river.my_field.get_available_states()
        self.assertEqual(1, available_states.count())
        self.assertEqual(self.state2, available_states[0])

        available_states = object.river.my_field.get_available_states(as_user=self.user1)
        self.assertEqual(1, available_states.count())
        self.assertEqual(self.state2, available_states[0])

        available_states = object.river.my_field.get_available_states(as_user=self.user2)
        self.assertEqual(0, available_states.count())

        available_states = object.river.my_field.get_available_states(as_user=self.user3)
        self.assertEqual(0, available_states.count())

        available_states = object.river.my_field.get_available_states(as_user=self.user4)
        self.assertEqual(0, available_states.count())

    def test_get_initial_state(self):
        self.initialize_standard_scenario()
        self.assertEqual(self.state1, TestModel.river.my_field.initial_state)

    def test_get_final_states(self):
        self.initialize_standard_scenario()
        self.assertListEqual([self.state41, self.state42, self.state51, self.state52], list(TestModel.river.my_field.final_states))

    def test_get_waiting_transition_approvals_without_skip(self):
        self.initialize_standard_scenario()
        object = TestModelObjectFactory.create_batch(1)[0]

        transition_approvals = object.river.my_field.get_available_approvals(as_user=self.user1)
        self.assertEqual(1, transition_approvals.count())

        transition_approvals = object.river.my_field.get_available_approvals(as_user=self.user2)
        self.assertEqual(0, transition_approvals.count())

        transition_approvals = object.river.my_field.get_available_approvals(as_user=self.user3)
        self.assertEqual(0, transition_approvals.count())

        transition_approvals = object.river.my_field.get_available_approvals(as_user=self.user4)
        self.assertEqual(0, transition_approvals.count())

    def test_get_waiting_transition_approvals_with_skip(self):
        self.initialize_standard_scenario()
        object = TestModelObjectFactory.create_batch(1)[0]

        transition_approvals = object.river.my_field.get_available_approvals(as_user=self.user1)
        self.assertEqual(1, transition_approvals.count())
        self.assertEqual(self.state2, transition_approvals[0].destination_state)

        TransitionApproval.objects.filter(
            workflow_object=object,
            field_name="my_field",
            destination_state=self.state2
        ).update(skip=True)

        transition_approvals = object.river.my_field.get_available_approvals(as_user=self.user2)
        self.assertEqual(1, transition_approvals.count())
        self.assertEqual(self.state3, transition_approvals[0].destination_state)

        TransitionApproval.objects.filter(
            workflow_object=object,
            field_name="my_field",
            destination_state=self.state3
        ).update(skip=True)

        transition_approvals = object.river.my_field.get_available_approvals(as_user=self.user4)
        self.assertEqual(2, transition_approvals.count())
        self.assertEqual(self.state4, transition_approvals[0].destination_state)
        self.assertEqual(self.state5, transition_approvals[1].destination_state)

        TransitionApproval.objects.filter(
            workflow_object=object,
            field_name="my_field",
            destination_state=self.state4
        ).update(skip=True)

        transition_approvals = object.river.my_field.get_available_approvals(as_user=self.user4)
        self.assertEqual(3, transition_approvals.count())
        self.assertEqual(self.state5, transition_approvals[0].destination_state)
        self.assertEqual(self.state41, transition_approvals[1].destination_state)
        self.assertEqual(self.state42, transition_approvals[2].destination_state)

        TransitionApproval.objects.filter(
            workflow_object=object,
            field_name="my_field",
            destination_state=self.state4
        ).update(skip=False)

        TransitionApproval.objects.filter(
            workflow_object=object,
            field_name="my_field",
            destination_state=self.state5
        ).update(skip=True)

        transition_approvals = object.river.my_field.get_available_approvals(as_user=self.user4)
        self.assertEqual(3, transition_approvals.count())
        self.assertEqual(self.state4, transition_approvals[0].destination_state)
        self.assertEqual(self.state51, transition_approvals[1].destination_state)
        self.assertEqual(self.state52, transition_approvals[2].destination_state)

        TransitionApproval.objects.filter(
            workflow_object=object,
            field_name="my_field",
            destination_state__in=[self.state4, self.state5]
        ).update(skip=True)

        transition_approvals = object.river.my_field.get_available_approvals(as_user=self.user4)
        self.assertEqual(4, transition_approvals.count())
        self.assertEqual(self.state41, transition_approvals[0].destination_state)
        self.assertEqual(self.state42, transition_approvals[1].destination_state)
        self.assertEqual(self.state51, transition_approvals[2].destination_state)
        self.assertEqual(self.state52, transition_approvals[3].destination_state)

        TransitionApproval.objects.filter(
            workflow_object=object,
            field_name="my_field",
            destination_state__in=[self.state41, self.state51]
        ).update(skip=True)

        transition_approvals = object.river.my_field.get_available_approvals(as_user=self.user4)
        self.assertEqual(2, transition_approvals.count())
        self.assertEqual(self.state42, transition_approvals[0].destination_state)
        self.assertEqual(self.state52, transition_approvals[1].destination_state)

    def test_proceed(self):
        self.initialize_standard_scenario()
        object = TestModelObjectFactory.create_batch(1)[0]

        # ####################
        # STATE 1 - STATE 2
        # Only User1(2001) can proceed and after his proceed state must be changed to STATE 2
        # ###################

        # Proceeded by user has no required permission for this transition

        try:
            object.river.my_field.approve(as_user=self.user2)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)

        try:
            object.river.my_field.approve(as_user=self.user3)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)

        try:
            object.river.my_field.approve(as_user=self.user4)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)

        self.assertEqual(self.state1, object.my_field)

        object.river.my_field.approve(as_user=self.user1)

        self.assertEqual(self.state2, object.my_field)

        transition_approvals = TransitionApproval.objects.filter(
            workflow_object=object,
            status=APPROVED,
            source_state=self.state1,
            destination_state=self.state2
        )
        self.assertEqual(1, transition_approvals.count())
        self.assertIsNotNone(transition_approvals[0].transactioner)
        self.assertEqual(self.user1, transition_approvals[0].transactioner)
        self.assertIsNotNone(transition_approvals[0].transaction_date)

        # ####################
        # STATE 2 - STATE 3
        # User2(2002) is first proceeder and User3(2003) is second proceeder. This must be done with turn. After proceeding is done, state is gonna be changed to STATE 3
        # ####################

        # Proceeded by user has no required permission for this transition
        try:
            object.river.my_field.approve(as_user=self.user1)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)

        try:
            object.river.my_field.approve(as_user=self.user4)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)

        # Turn is User2(2002)s, not User3(2003)s. After User2(2002) proceeded, User3(2003) can proceed.
        try:
            object.river.my_field.approve(as_user=self.user3)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)

        # Proceeded by two user has required permission for this transition to get next state (order is user2(2002),user3(2003)).
        object.river.my_field.approve(as_user=self.user2)
        self.assertEqual(self.state2, object.my_field)

        transition_approvals = TransitionApproval.objects.filter(
            workflow_object=object,
            status=APPROVED,
            source_state=self.state2,
            destination_state=self.state3
        )
        self.assertEqual(1, transition_approvals.count())
        self.assertIsNotNone(transition_approvals[0].transactioner)
        self.assertEqual(self.user2, transition_approvals[0].transactioner)
        self.assertIsNotNone(transition_approvals[0].transaction_date)

        try:
            object.river.my_field.approve(as_user=self.user2)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)

            object.river.my_field.approve(as_user=self.user3)
        self.assertEqual(self.state3, object.my_field)

        transition_approvals = TransitionApproval.objects.filter(
            workflow_object=object,
            status=APPROVED,
            source_state=self.state2,
            destination_state=self.state3
        )
        self.assertEqual(2, transition_approvals.count())
        self.assertIsNotNone(transition_approvals[0].transactioner)
        self.assertIsNotNone(transition_approvals[1].transactioner)
        self.assertEqual(self.user2, transition_approvals[0].transactioner)
        self.assertEqual(self.user3, transition_approvals[1].transactioner)
        self.assertIsNotNone(transition_approvals[0].transaction_date)
        self.assertIsNotNone(transition_approvals[1].transaction_date)

        # ####################
        # STATE 3 - STATE 4 or STATE 5
        # Only User4(2004) can proceed by giving the exact next state and after his proceed with his state must be changed to STATE 4 or STATE 5
        # ###################

        # Proceeded by user has no required permission for this transition
        try:
            object.river.my_field.approve(as_user=self.user1)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)

        # Proceeded by user has no required permission for this transition
        try:
            object.river.my_field.approve(as_user=self.user2)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)

        # Proceeded by user has no required permission for this transition
        try:
            object.river.my_field.approve(as_user=self.user3)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)

        # There are STATE 4 and STATE 5 as next. State must be given to switch
        try:
            object.river.my_field.approve(as_user=self.user4)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'State must be given when there are multiple states for destination')
            self.assertEqual(ErrorCode.NEXT_STATE_IS_REQUIRED, e.code)

        # There are STATE 4 and STATE 5 as next. State among STATE 4 and STATE 5 must be given to switch, not other state
        try:
            object.river.my_field.approve(as_user=self.user4, next_state=self.state3)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e),
                             "Invalid state is given(%s). Valid states is(are) %s" % (
                                 self.state3.__str__(), ','.join([ast.__str__() for ast in [self.state4, self.state5]])))
            self.assertEqual(ErrorCode.INVALID_NEXT_STATE_FOR_USER, e.code)

        # There are STATE 4 and STATE 5 as next. After one of them is given to proceeding, the state must be switch to it immediately.
        object.river.my_field.approve(as_user=self.user4, next_state=self.state5)
        self.assertEqual(self.state5, object.my_field)

        transition_approvals = TransitionApproval.objects.filter(
            workflow_object=object,
            status=APPROVED,
            source_state=self.state3,
            destination_state=self.state5
        )
        self.assertEqual(1, transition_approvals.count())
        self.assertIsNotNone(transition_approvals[0].transactioner)
        self.assertEqual(self.user4, transition_approvals[0].transactioner)
        self.assertIsNotNone(transition_approvals[0].transaction_date)

    def test_cycle_proceedings(self):
        self.initialize_circular_scenario()
        object = TestModelObjectFactory.create_batch(1)[0]

        # No Cycle
        self.assertFalse(object.river.my_field._cycle_proceedings())
        object.river.my_field.approve(as_user=self.user1, next_state=self.in_progress_state, god_mod=True)
        self.assertEqual(5, TransitionApproval.objects.filter(object_id=object.pk).count())

        # No Cycle
        self.assertFalse(object.river.my_field._cycle_proceedings())
        object.river.my_field.approve(as_user=self.user2, next_state=self.resolved_state, god_mod=True)
        self.assertEqual(5, TransitionApproval.objects.filter(object_id=object.pk).count())

        # State is re-opened and cycle is detected. Transition in-progress to resolved proceeding is cloned
        self.assertFalse(object.river.my_field._cycle_proceedings())
        object.river.my_field.approve(as_user=self.user3, next_state=self.re_opened_state, god_mod=True)
        self.assertEqual(6, TransitionApproval.objects.filter(object_id=object.pk).count())

        self.assertEqual(TransitionApprovalMeta.objects.get(source_state=self.in_progress_state, destination_state=self.resolved_state),
                         TransitionApproval.objects.filter(object_id=object.pk).latest('date_created').meta)

        # There will be no cycling even if the method is invoked. Because cycling is done in proceeding.
        self.assertFalse(object.river.my_field._cycle_proceedings())
        self.assertEqual(6, TransitionApproval.objects.filter(object_id=object.pk).count())

        # State is in-progress and cycle is detected. Transition resolved to re-opened proceeding is cloned
        object.river.my_field.approve(as_user=self.user3, next_state=self.in_progress_state, god_mod=True)
        self.assertEqual(7, TransitionApproval.objects.filter(object_id=object.pk).count())
        self.assertEqual(TransitionApprovalMeta.objects.get(source_state=self.resolved_state, destination_state=self.re_opened_state),
                         TransitionApproval.objects.filter(object_id=object.pk).latest('date_created').meta)

        # State is resolved and cycle is detected. Transition re-opened to in-progress proceeding is cloned
        object.river.my_field.approve(as_user=self.user3, next_state=self.resolved_state, god_mod=True)
        self.assertEqual(8, TransitionApproval.objects.filter(object_id=object.pk).count())
        self.assertEqual(TransitionApprovalMeta.objects.get(source_state=self.re_opened_state, destination_state=self.in_progress_state),
                         TransitionApproval.objects.filter(object_id=object.pk).latest('date_created').meta)

        # State is re-opened and cycle is detected. Transition  in-progress to resolved proceeding is cloned
        self.assertFalse(object.river.my_field._cycle_proceedings())
        object.river.my_field.approve(as_user=self.user3, next_state=self.re_opened_state, god_mod=True)
        self.assertEqual(9, TransitionApproval.objects.filter(object_id=object.pk).count())
        self.assertEqual(TransitionApprovalMeta.objects.get(source_state=self.in_progress_state, destination_state=self.resolved_state),
                         TransitionApproval.objects.filter(object_id=object.pk).latest('date_created').meta)

        # State is in-progress and cycle is detected. Transition resolved to re-opened proceeding is cloned
        object.river.my_field.approve(as_user=self.user3, next_state=self.in_progress_state, god_mod=True)
        self.assertEqual(10, TransitionApproval.objects.filter(object_id=object.pk).count())
        self.assertEqual(TransitionApprovalMeta.objects.get(source_state=self.resolved_state, destination_state=self.re_opened_state),
                         TransitionApproval.objects.filter(object_id=object.pk).latest('date_created').meta)

        # State is resolved and cycle is detected. Transition re-opened to in-progress proceeding is cloned
        object.river.my_field.approve(as_user=self.user3, next_state=self.resolved_state, god_mod=True)
        self.assertEqual(11, TransitionApproval.objects.filter(object_id=object.pk).count())
        self.assertEqual(TransitionApprovalMeta.objects.get(source_state=self.re_opened_state, destination_state=self.in_progress_state),
                         TransitionApproval.objects.filter(object_id=object.pk).latest('date_created').meta)

        # No Cycle for closed state.
        object.river.my_field.approve(as_user=self.user4, next_state=self.closed_state, god_mod=True)
        self.assertEqual(11, TransitionApproval.objects.filter(object_id=object.pk).count())

    def test_get_waiting_approvals_slowness_test(self):
        self.initialize_standard_scenario()
        self.objects = TestModelObjectFactory.create_batch(100)
        before = datetime.now()
        for o in self.objects:
            o.river.my_field.get_available_states(as_user=self.user1)
        after = datetime.now()
        self.assertLess(after - before, timedelta(seconds=2))



from django import forms
from django.contrib import admin
from django.contrib.contenttypes.models import ContentType

from river.models.fields.state import StateField
from river.core.workflowregistry import workflow_registry
from river.models.transitionapprovalmeta import TransitionApprovalMeta


def get_workflow_choices():
    class_by_id = lambda cid: workflow_registry.class_index[cid]
    result = []
    for class_id, field_names in workflow_registry.workflows.items():
        cls = class_by_id(class_id)
        content_type = ContentType.objects.get_for_model(cls)
        for field_name in field_names:
            result.append(("%s %s" % (content_type.pk, field_name), "%s.%s - %s" % (cls.__module__, cls.__name__, field_name)))
    return result


class TransitionApprovalMetaForm(forms.ModelForm):
    workflow = forms.ChoiceField(choices=[])

    class Meta:
        model = TransitionApprovalMeta
        fields = ('workflow', 'source_state', 'destination_state', 'permissions', 'groups', 'priority', 'action_text')

    def __init__(self, *args, **kwargs):
        instance = kwargs.get("instance", None)
        self.declared_fields['workflow'].choices = get_workflow_choices()
        if instance and instance.pk:
            self.declared_fields['workflow'].initial = "%s %s" % (instance.content_type.pk, instance.field_name)

        super(TransitionApprovalMetaForm, self).__init__(*args, **kwargs)

    def clean_workflow(self):
        if self.cleaned_data.get('workflow') == '' or ' ' not in self.cleaned_data.get('workflow'):
            return None, None
        else:
            return self.cleaned_data.get('workflow').split(" ")

    def save(self, *args, **kwargs):
        content_type_pk, field_name = self.cleaned_data.get('workflow')
        instance = super(TransitionApprovalMetaForm, self).save(commit=False)
        instance.content_type = ContentType.objects.get(pk=content_type_pk)
        instance.field_name = field_name
        return super(TransitionApprovalMetaForm, self).save(*args, **kwargs)


class TransitionApprovalMetaAdmin(admin.ModelAdmin):
    form = TransitionApprovalMetaForm
    list_display = ('model_class', 'field_name', 'source_state', 'destination_state', 'priority')

    def model_class(self, obj):
        cls = obj.content_type.model_class()
        if cls:
            return "%s.%s" % (cls.__module__, cls.__name__)
        else:
            return "Class not found in the workspace"


admin.site.register(TransitionApprovalMeta, TransitionApprovalMetaAdmin)



from river.hooking.completed import PostCompletedHooking
from river.tests.base_test import BaseTestCase
from river.tests.models.factories import TestModelObjectFactory

__author__ = 'ahmetdal'


class CompletedHookingTest(BaseTestCase):
    def setUp(self):
        super(CompletedHookingTest, self).setUp()
        self.initialize_standard_scenario()

    def test_register_for_an_object(self):
        objects = TestModelObjectFactory.create_batch(2)

        self.test_args = None
        self.test_kwargs = None

        def test_callback(*args, **kwargs):
            self.test_args = args
            self.test_kwargs = kwargs

        objects[0].river.my_field.hook_post_complete(test_callback)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        objects[0].river.my_field.approve(as_user=self.user1)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        # Proceeded but no transition
        objects[0].river.my_field.approve(as_user=self.user2)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        objects[0].river.my_field.approve(as_user=self.user3)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        objects[0].river.my_field.approve(as_user=self.user4, next_state=self.state4)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        objects[0].river.my_field.approve(as_user=self.user4, next_state=self.state41)

        self.assertEqual((objects[0], "my_field"), self.test_args)



from __future__ import unicode_literals

from django.db import models
from django.db.models.signals import pre_save
from django.template.defaultfilters import slugify
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

from river.models.base_model import BaseModel
from river.models.managers.state import StateManager

__author__ = 'ahmetdal'


@python_2_unicode_compatible
class State(BaseModel):
    class Meta:
        app_label = 'river'
        verbose_name = _("State")
        verbose_name_plural = _("States")

    objects = StateManager()

    slug = models.SlugField(unique=True, null=True, blank=True)
    label = models.CharField(max_length=50)
    description = models.CharField(_("Description"), max_length=200, null=True, blank=True)

    def __str__(self):
        return self.label

    def natural_key(self):
        return self.slug,

    def details(self):
        detail = super(State, self).details()
        detail.update(
            {
                'label': self.label,
                'description': self.description
            }
        )
        return detail


def on_pre_save(sender, instance, *args, **kwargs):
    if not instance.slug:
        instance.slug = slugify(instance.label)
    else:
        # To be sure it is slug
        instance.slug = slugify(instance.slug)


pre_save.connect(on_pre_save, State)



from django.db.models import CASCADE
from mptt.fields import TreeOneToOneField

from river.models import State, TransitionApprovalMeta

try:
    from django.contrib.contenttypes.fields import GenericForeignKey
except ImportError:
    from django.contrib.contenttypes.generic import GenericForeignKey

from django.db import models
from django.utils.translation import ugettext_lazy as _

from river.models.base_model import BaseModel
from river.models.managers.transitionapproval import TransitionApprovalManager
from river.config import app_config

__author__ = 'ahmetdal'

PENDING = 0
APPROVED = 1
REJECTED = 2

PROCEEDING_STATUSES = [
    (PENDING, _('Pending')),
    (APPROVED, _('Approved')),
    (REJECTED, _('Rejected')),
]


class TransitionApproval(BaseModel):
    class Meta:
        app_label = 'river'
        verbose_name = _("Transition Approval")
        verbose_name_plural = _("Transition Approvals")

    objects = TransitionApprovalManager()

    content_type = models.ForeignKey(app_config.CONTENT_TYPE_CLASS, verbose_name=_('Content Type'), on_delete=CASCADE)
    field_name = models.CharField(_("Field Name"), max_length=200)

    object_id = models.CharField(max_length=50, verbose_name=_('Related Object'))
    workflow_object = GenericForeignKey('content_type', 'object_id')

    meta = models.ForeignKey(TransitionApprovalMeta, verbose_name=_('Meta'), related_name="proceedings", on_delete=CASCADE)
    source_state = models.ForeignKey(State, verbose_name=_("Source State"), related_name='transition_approvals_as_source', on_delete=CASCADE)
    destination_state = models.ForeignKey(State, verbose_name=_("Next State"), related_name='transition_approvals_as_destination', on_delete=CASCADE)

    transactioner = models.ForeignKey(app_config.USER_CLASS, verbose_name=_('Transactioner'), null=True, blank=True, on_delete=CASCADE)
    transaction_date = models.DateTimeField(null=True, blank=True)

    status = models.IntegerField(_('Status'), choices=PROCEEDING_STATUSES, default=PENDING)

    skip = models.BooleanField(_('Skip'), default=False)

    permissions = models.ManyToManyField(app_config.PERMISSION_CLASS, verbose_name=_('Permissions'))
    groups = models.ManyToManyField(app_config.GROUP_CLASS, verbose_name=_('Groups'))
    priority = models.IntegerField(default=0, verbose_name=_('Priority'))

    enabled = models.BooleanField(_('Enabled?'), default=True)

    previous = TreeOneToOneField("self", verbose_name=_('Previous Transition'), related_name="next_transition", null=True, blank=True, on_delete=CASCADE)

    cloned = models.BooleanField(_('Cloned?'), default=False)



from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

from river.models.factories import StateObjectFactory, TransitionApprovalMetaFactory, PermissionObjectFactory, UserObjectFactory
from river.tests.models import TestModel

__author__ = 'ahmetdal'


class BaseTestCase(TestCase):
    def initialize_standard_scenario(self):
        TransitionApprovalMetaFactory.reset_sequence(0)
        StateObjectFactory.reset_sequence(0)

        content_type = ContentType.objects.get_for_model(TestModel)
        permissions = PermissionObjectFactory.create_batch(4)
        self.user1 = UserObjectFactory(user_permissions=[permissions[0]])
        self.user2 = UserObjectFactory(user_permissions=[permissions[1]])
        self.user3 = UserObjectFactory(user_permissions=[permissions[2]])
        self.user4 = UserObjectFactory(user_permissions=[permissions[3]])

        self.state1 = StateObjectFactory(label="state1")
        self.state2 = StateObjectFactory(label="state2")
        self.state3 = StateObjectFactory(label="state3")
        self.state4 = StateObjectFactory(label="state4")
        self.state5 = StateObjectFactory(label="state5")

        self.state41 = StateObjectFactory(label="state4.1")
        self.state42 = StateObjectFactory(label="state4.2")

        self.state51 = StateObjectFactory(label="state5.1")
        self.state52 = StateObjectFactory(label="state5.2")

        t1 = TransitionApprovalMetaFactory.create(
            field_name="my_field",
            content_type=content_type,
            source_state=self.state1,
            destination_state=self.state2,
            priority=0
        )
        t1.permissions.add(permissions[0])

        t2 = TransitionApprovalMetaFactory.create(
            field_name="my_field",
            content_type=content_type,
            source_state=self.state2,
            destination_state=self.state3,
            priority=0
        )
        t2.permissions.add(permissions[1])

        t3 = TransitionApprovalMetaFactory.create(
            field_name="my_field",
            content_type=content_type,
            source_state=self.state2,
            destination_state=self.state3,
            priority=1
        )
        t3.permissions.add(permissions[2])

        t4 = TransitionApprovalMetaFactory.create(
            field_name="my_field",
            content_type=content_type,
            source_state=self.state3,
            destination_state=self.state4,
            priority=0
        )
        t4.permissions.add(permissions[3])

        t5 = TransitionApprovalMetaFactory.create(
            field_name="my_field",
            content_type=content_type,
            source_state=self.state3,
            destination_state=self.state5,
            priority=0
        )
        t5.permissions.add(permissions[3])

        t6 = TransitionApprovalMetaFactory.create(
            field_name="my_field",
            content_type=content_type,
            source_state=self.state4,
            destination_state=self.state41,
            priority=0
        )
        t6.permissions.add(permissions[3])

        t7 = TransitionApprovalMetaFactory.create(
            field_name="my_field",
            content_type=content_type,
            source_state=self.state4,
            destination_state=self.state42,
            priority=0
        )
        t7.permissions.add(permissions[3])

        t8 = TransitionApprovalMetaFactory.create(
            field_name="my_field",
            content_type=content_type,
            source_state=self.state5,
            destination_state=self.state51,
            priority=0
        )
        t8.permissions.add(permissions[3])

        t9 = TransitionApprovalMetaFactory.create(
            field_name="my_field",
            content_type=content_type,
            source_state=self.state5,
            destination_state=self.state52,
            priority=0
        )
        t9.permissions.add(permissions[3])

    def initialize_circular_scenario(self):
        StateObjectFactory.reset_sequence(0)
        TransitionApprovalMetaFactory.reset_sequence(0)

        content_type = ContentType.objects.get_for_model(TestModel)
        permissions = PermissionObjectFactory.create_batch(4)
        self.user1 = UserObjectFactory(user_permissions=[permissions[0]])
        self.user2 = UserObjectFactory(user_permissions=[permissions[1]])
        self.user3 = UserObjectFactory(user_permissions=[permissions[2]])
        self.user4 = UserObjectFactory(user_permissions=[permissions[3]])

        self.open_state = StateObjectFactory(label='open')
        self.in_progress_state = StateObjectFactory(label='in-progress')
        self.resolved_state = StateObjectFactory(label='resolved')
        self.re_opened_state = StateObjectFactory(label='re-opened')
        self.closed_state = StateObjectFactory(label='closed')

        t1 = TransitionApprovalMetaFactory.create(
            field_name="my_field",
            content_type=content_type,
            source_state=self.open_state,
            destination_state=self.in_progress_state,
        )
        t1.permissions.add(permissions[0])

        t2 = TransitionApprovalMetaFactory.create(
            field_name="my_field",
            content_type=content_type,
            source_state=self.in_progress_state,
            destination_state=self.resolved_state,
        )
        t2.permissions.add(permissions[1])

        t3 = TransitionApprovalMetaFactory.create(
            field_name="my_field",
            content_type=content_type,
            source_state=self.resolved_state,
            destination_state=self.re_opened_state,
        )
        t3.permissions.add(permissions[2])

        t4 = TransitionApprovalMetaFactory.create(
            field_name="my_field",
            content_type=content_type,
            source_state=self.resolved_state,
            destination_state=self.closed_state,
        )
        t4.permissions.add(permissions[3])

        t5 = TransitionApprovalMetaFactory.create(
            field_name="my_field",
            content_type=content_type,
            source_state=self.re_opened_state,
            destination_state=self.in_progress_state,
        )
        t5.permissions.add(permissions[0])



from river.config import app_config
from river.hooking.backends.loader import load_callback_backend
from river.hooking.transition import PostTransitionHooking
from river.tests.base_test import BaseTestCase
from river.tests.models.factories import TestModelObjectFactory

__author__ = 'ahmetdal'


def test_handler(*args, **kwargs):
    pass


class MemoryHookingBackendTest(BaseTestCase):
    def setUp(self):
        super(MemoryHookingBackendTest, self).setUp()
        self.field_name = "my_field"
        self.initialize_standard_scenario()
        app_config.HOOKING_BACKEND_CLASS = 'river.hooking.backends.memory.MemoryHookingBackend'
        self.handler_backend = load_callback_backend()
        self.handler_backend.callbacks = {}

    def test_register(self):
        objects = TestModelObjectFactory.create_batch(2)

        self.assertFalse('%s.%s_object%s_field_name%s' % (
            PostTransitionHooking.__module__, PostTransitionHooking.__name__, objects[1].pk, self.field_name) in self.handler_backend.callbacks)
        self.handler_backend.register(PostTransitionHooking, test_handler, objects[1], self.field_name)
        self.assertTrue(
            '%s.%s_object%s_field_name%s' % (PostTransitionHooking.__module__, PostTransitionHooking.__name__, objects[1].pk, self.field_name) in self.handler_backend.callbacks)
        self.assertEqual(test_handler.__name__, list(self.handler_backend.callbacks.values())[0].__name__)

    def test_get_handlers(self):
        object = TestModelObjectFactory.create_batch(1)[0]

        self.handler_backend.register(PostTransitionHooking, test_handler, object, self.field_name)
        callbacks = self.handler_backend.get_callbacks(PostTransitionHooking, object, self.field_name)
        self.assertEqual(1, len(callbacks))

        self.assertEqual(test_handler.__name__, callbacks[0].__name__)



from django.contrib.contenttypes.models import ContentType

from river.models import State, TransitionApprovalMeta
from river.utils.error_code import ErrorCode
from river.utils.exceptions import RiverException


class ClassWorkflowObject(object):

    def __init__(self, workflow_class, name, field_name):
        self.workflow_class = workflow_class
        self.name = name
        self.field_name = field_name

    @property
    def _content_type(self):
        return ContentType.objects.get_for_model(self.workflow_class)

    def get_on_approval_objects(self, as_user):
        object_pks = []
        for workflow_object in self.workflow_class.objects.all():
            instance_workflow = getattr(workflow_object.river, self.name)
            transition_approvals = instance_workflow.get_available_approvals(as_user=as_user)
            if transition_approvals.count():
                object_pks.append(workflow_object.pk)
        return self.workflow_class.objects.filter(pk__in=object_pks)

    @property
    def initial_state(self):
        initial_states = State.objects.filter(
            pk__in=TransitionApprovalMeta.objects.filter(
                content_type=self._content_type,
                field_name=self.name,
                parents__isnull=True
            ).values_list("source_state", flat=True)
        )
        if initial_states.count() == 0:
            raise RiverException(ErrorCode.NO_AVAILABLE_INITIAL_STATE, 'There is no available initial state for the content type %s. ' % self._content_type)
        elif initial_states.count() > 1:
            raise RiverException(ErrorCode.MULTIPLE_INITIAL_STATE,
                                 'There are multiple initial state for the content type %s. Have only one initial state' % self._content_type)

        return initial_states[0]

    @property
    def final_states(self):
        return State.objects.filter(
            pk__in=TransitionApprovalMeta.objects.filter(
                field_name=self.name,
                children__isnull=True,
                content_type=self._content_type
            ).values_list("destination_state", flat=True)
        )


import logging

from django.contrib import auth
from django.contrib.contenttypes.models import ContentType
from django.db.models import Min, Q
from django.db.transaction import atomic
from django.utils import timezone

from river.config import app_config
from river.hooking.completed import PostCompletedHooking, PreCompletedHooking
from river.hooking.transition import PostTransitionHooking, PreTransitionHooking
from river.models import TransitionApproval, TransitionApprovalMeta, PENDING, State, APPROVED
from river.signals import ProceedingSignal, TransitionSignal, FinalSignal
from river.utils.error_code import ErrorCode
from river.utils.exceptions import RiverException

LOGGER = logging.getLogger(__name__)


class InstanceWorkflowObject(object):

    def __init__(self, workflow_object, name, field_name):
        self.class_workflow = getattr(workflow_object.__class__.river, name)
        self.workflow_object = workflow_object
        self.name = name
        self.field_name = field_name
        self.initialized = False

    def initialize_approvals(self):
        if not self.initialized and not TransitionApproval.objects.filter(workflow_object=self.workflow_object, field_name=self.name).count():
            content_type = app_config.CONTENT_TYPE_CLASS.objects.get_for_model(self.workflow_object)
            for transition_approval_meta in TransitionApprovalMeta.objects.filter(field_name=self.name, content_type=content_type):
                transition_approval, created = TransitionApproval.objects.update_or_create(
                    workflow_object=self.workflow_object,
                    field_name=transition_approval_meta.field_name,
                    source_state=transition_approval_meta.source_state,
                    destination_state=transition_approval_meta.destination_state,
                    priority=transition_approval_meta.priority,
                    meta=transition_approval_meta,
                    defaults={
                        'status': PENDING,
                    }
                )
                transition_approval.permissions.add(*transition_approval_meta.permissions.all())
                transition_approval.groups.add(*transition_approval_meta.groups.all())
            # self.workflow_object.save()
            self.initialized = True
            LOGGER.debug("Transition approvals are initialized for the workflow object %s" % self.workflow_object)

    @property
    def on_initial_state(self):
        return self.get_state() == self.class_workflow.initial_state

    @property
    def on_final_state(self):
        if self.class_workflow.final_states.count() == 0:
            raise RiverException(ErrorCode.NO_AVAILABLE_FINAL_STATE, 'There is no available final state for the content type %s.' % self._content_type)

        return self.get_state() in self.class_workflow.final_states

    @property
    def next_approvals(self):
        return self._get_next_approvals()

    @property
    def recent_approval(self):
        try:

            return getattr(self.workflow_object, self.name + "_transitions").filter(transaction_date__isnull=False).latest('transaction_date')
        except TransitionApproval.DoesNotExist:
            return None

    @property
    def _content_type(self):
        return ContentType.objects.get_for_model(self.workflow_object)

    # def get_available_transitions(self, as_user, *args, **kwargs):
    # from river.services.proceeding import ProceedingService
    # return ProceedingService.get_available_proceedings(self.workflow_object, [self.get_state()], user=as_user, *args, **kwargs)

    def get_available_states(self, as_user=None):
        transition_approvals = TransitionApproval.objects.filter(
            workflow_object=self.workflow_object,
            source_state=self.get_state(),
        )
        if as_user:
            transition_approvals = transition_approvals.filter(
                permissions__in=as_user.user_permissions.all()
            )
        destination_states = transition_approvals.values_list('destination_state', flat=True)
        return State.objects.filter(pk__in=destination_states)

    def get_available_approvals(self, as_user=None, source_states=None, destination_state=None, god_mod=False):
        def get_transition_approvals(transition_approvals):
            min_priority = transition_approvals.aggregate(Min('priority'))['priority__min']
            transition_approvals = transition_approvals.filter(priority=min_priority)

            if destination_state:
                transition_approvals = transition_approvals.filter(destination_state=destination_state)

            return transition_approvals

        def authorize_transition_approvals(transition_approvals):
            group_q = Q()
            for g in as_user.groups.all():
                group_q = group_q | Q(groups__in=[g])

            permissions = []
            for backend in auth.get_backends():
                permissions.extend(backend.get_all_permissions(as_user))

            permission_q = Q()
            for p in permissions:
                label, codename = p.split('.')
                permission_q = permission_q | Q(permissions__content_type__app_label=label,
                                                permissions__codename=codename)

            return transition_approvals.filter(
                (
                        (Q(transactioner__isnull=True) | Q(transactioner=as_user)) &
                        (Q(permissions__isnull=True) | permission_q) &
                        (Q(groups__isnull=True) | group_q)
                )
            )

        source_states = source_states or [self.get_state()]

        transition_approvals = TransitionApproval.objects.filter(
            workflow_object=self.workflow_object,
            field_name=self.name,
            source_state__in=source_states,
            status=PENDING,
            enabled=True
        )

        suitable_transition_approvals = get_transition_approvals(transition_approvals.filter(skip=False))

        if as_user and not god_mod:
            suitable_transition_approvals = authorize_transition_approvals(suitable_transition_approvals)

        skipped_transition_approvals = get_transition_approvals(transition_approvals.filter(skip=True))
        if skipped_transition_approvals:
            source_state_pks = list(skipped_transition_approvals.values_list('destination_state', flat=True))
            suitable_transition_approvals = suitable_transition_approvals | self.get_available_approvals(
                as_user=as_user,
                source_states=State.objects.filter(pk__in=source_state_pks),
                destination_state=destination_state,
                god_mod=god_mod
            )
        return suitable_transition_approvals

    @atomic
    def approve(self, as_user, next_state=None, god_mod=False):
        def process(action, next_state=None, god_mod=False):
            available_transition_approvals = self.get_available_approvals(as_user=as_user, god_mod=god_mod)
            c = available_transition_approvals.count()
            if c == 0:
                raise RiverException(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, "There is no available state for destination for the user.")
            if c > 1:
                if next_state:
                    available_transition_approvals = available_transition_approvals.filter(destination_state=next_state)
                    if available_transition_approvals.count() == 0:
                        available_states = self.get_available_states(as_user=as_user)
                        raise RiverException(ErrorCode.INVALID_NEXT_STATE_FOR_USER, "Invalid state is given(%s). Valid states is(are) %s" % (
                            next_state.__str__(), ','.join([ast.__str__() for ast in available_states])))
                else:
                    raise RiverException(ErrorCode.NEXT_STATE_IS_REQUIRED,
                                         "State must be given when there are multiple states for destination")

            available_transition_approval = available_transition_approvals[0]
            available_transition_approval.status = action
            available_transition_approval.transactioner = as_user
            available_transition_approval.transaction_date = timezone.now()
            if self.recent_approval:
                available_transition_approval.previous = self.recent_approval
            available_transition_approval.save()

            return available_transition_approval

        transition_approval = process(APPROVED, next_state, god_mod)

        # Any other transitions approval is left?
        rest_of_transition_approvals = self.get_available_approvals(destination_state=next_state, god_mod=True)

        transition_status = False
        previous_state = self.get_state()
        if rest_of_transition_approvals.count() == 0:
            self.set_state(transition_approval.destination_state)
            transition_status = True

            # Next states should be PENDING back again if there is circle.
            self._cycle_proceedings()
            # ProceedingService.get_next_proceedings(workflow_object).update(status=PENDING)

        with ProceedingSignal(self.workflow_object, self.field_name, transition_approval), \
             TransitionSignal(transition_status, self.workflow_object, self.field_name, transition_approval), \
             FinalSignal(self.workflow_object, self.field_name):
            self.workflow_object.save()

        LOGGER.debug("Workflow object %s is proceeded for next transition. Transition: %s -> %s" % (
            self.workflow_object, previous_state, self.get_state()))

    def _get_next_approvals(self, transition_approval_pks=None, current_states=None, index=0, limit=None):
        if not transition_approval_pks:
            transition_approval_pks = []
        index += 1
        current_states = list(current_states.values_list('pk', flat=True)) if current_states else [self.get_state()]
        next_approvals = TransitionApproval.objects.filter(
            workflow_object=self.workflow_object,
            field_name=self.name,
            source_state__in=current_states
        )
        if self.recent_approval:
            next_approvals = next_approvals.exclude(pk=self.recent_approval.pk)
        if next_approvals.exists() and not next_approvals.filter(pk__in=transition_approval_pks).exists() and (
                not limit or index < limit):
            proceedings = self._get_next_approvals(
                transition_approval_pks=transition_approval_pks + list(next_approvals.values_list('pk', flat=True)),
                current_states=State.objects.filter(pk__in=next_approvals.values_list('destination_state', flat=True)),
                index=index,
                limit=limit
            )
        else:
            proceedings = TransitionApproval.objects.filter(pk__in=transition_approval_pks)

        return proceedings

    @atomic
    def _cycle_proceedings(self):
        """
         Finds next proceedings and clone them for cycling if it exists.
        """
        next_approvals = self._get_next_approvals().exclude(
            status=PENDING).exclude(cloned=True)
        for ta in next_approvals:
            clone_transition_approval, c = TransitionApproval.objects.get_or_create(
                source_state=ta.source_state,
                destination_state=ta.destination_state,
                content_type=ta.content_type,
                object_id=ta.object_id,
                field_name=ta.field_name,
                skip=ta.skip,
                priority=ta.priority,
                enabled=ta.enabled,
                status=PENDING,
                meta=ta.meta
            )

            if c:
                clone_transition_approval.permissions.add(*ta.permissions.all())
                clone_transition_approval.groups.add(*ta.groups.all())
            next_approvals.update(cloned=True)

        return True if next_approvals.count() else False

    def get_state(self):
        return getattr(self.workflow_object, self.field_name)

    def set_state(self, state):
        return setattr(self.workflow_object, self.field_name, state)

    def hook_post_transition(self, callback, *args, **kwargs):
        PostTransitionHooking.register(callback, self.workflow_object, self.field_name, *args, **kwargs)

    def hook_pre_transition(self, callback, *args, **kwargs):
        PreTransitionHooking.register(callback, self.workflow_object, self.field_name, *args, **kwargs)

    def hook_post_complete(self, callback):
        PostCompletedHooking.register(callback, self.workflow_object, self.field_name)

    def hook_pre_complete(self, callback):
        PreCompletedHooking.register(callback, self.workflow_object, self.field_name)



from django.contrib.auth.models import Permission

from river.models import TransitionApproval
from river.tests.base_test import BaseTestCase
from river.tests.models.factories import TestModelObjectFactory

__author__ = 'ahmetdal'


class TransitionHooking(BaseTestCase):

    def test_register_for_an_object(self):
        self.initialize_standard_scenario()
        self.test_args = None
        self.test_kwargs = None

        def test_callback(*args, **kwargs):
            self.test_args = args
            self.test_kwargs = kwargs

        objects = TestModelObjectFactory.create_batch(2)

        objects[1].river.my_field.hook_post_transition(test_callback)

        objects[0].river.my_field.approve(as_user=self.user1)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        objects[0].river.my_field.approve(as_user=self.user2)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        objects[0].river.my_field.hook_post_transition(test_callback)

        objects[0].river.my_field.approve(as_user=self.user3)

        self.assertEqual((objects[0], "my_field"), self.test_args)

        self.assertDictEqual(
            {
                'transition_approval': TransitionApproval.objects.get(object_id=objects[0].pk, source_state=self.state2, destination_state=self.state3,
                                                                      permissions__in=Permission.objects.filter(user=self.user3))
            }, self.test_kwargs)

    def test_register_for_a_transition(self):
        self.initialize_standard_scenario()

        self.test_args = None
        self.test_kwargs = None

        def test_callback(*args, **kwargs):
            self.test_args = args
            self.test_kwargs = kwargs

        objects = TestModelObjectFactory.create_batch(2)
        objects[0].river.my_field.hook_post_transition(test_callback, source_state=self.state2, destination_state=self.state3)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        objects[0].river.my_field.approve(as_user=self.user1)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        objects[0].river.my_field.approve(as_user=self.user2)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        objects[0].river.my_field.approve(as_user=self.user3)

        self.assertEqual((objects[0], "my_field"), self.test_args)
        self.assertDictEqual(
            {
                'transition_approval': TransitionApproval.objects.get(object_id=objects[0].pk, source_state=self.state2, destination_state=self.state3,
                                                                      permissions__in=Permission.objects.filter(user=self.user3))
            }, self.test_kwargs)



import logging

from django.dispatch import Signal

__author__ = 'ahmetdal'

pre_final = Signal(providing_args=["workflow_object", "field_name", ])
post_final = Signal(providing_args=["workflow_object", "field_name", ])

pre_transition = Signal(providing_args=["workflow_object", "field_name", "transition_approval"])
post_transition = Signal(providing_args=["workflow_object", "field_name", "transition_approval"])

pre_proceed = Signal(providing_args=["workflow_object", "field_name", "proceeding"])
post_proceed = Signal(providing_args=["workflow_object", "field_name", "proceeding"])

LOGGER = logging.getLogger(__name__)


class TransitionSignal(object):
    def __init__(self, status, workflow_object, field_name, transition_approval):
        self.status = status
        self.workflow_object = workflow_object
        self.field_name = field_name
        self.transition_approval = transition_approval

    def __enter__(self):
        if self.status:
            pre_transition.send(
                sender=TransitionSignal.__class__,
                workflow_object=self.workflow_object,
                field_name=self.field_name,
                source_state=self.transition_approval.source_state,
                destination_state=self.transition_approval.destination_state,
                transition_approval=self.transition_approval,
            )
            LOGGER.debug("Pre transition signal IS sent for workflow object %s for transition %s -> %s" % (
                self.workflow_object, self.transition_approval.source_state.label, self.transition_approval.destination_state.label))
        else:
            LOGGER.debug(
                "Pre transition signal IS NOT sent for workflow object %s. Although transition_approval is occurred, not transition is occurred." % (
                    self.workflow_object))

    def __exit__(self, type, value, traceback):
        if self.status:
            post_transition.send(
                sender=TransitionSignal.__class__,
                workflow_object=self.workflow_object,
                field_name=self.field_name,
                source_state=self.transition_approval.source_state,
                destination_state=self.transition_approval.destination_state,
                transition_approval=self.transition_approval,
            )
            LOGGER.debug(
                "Post transition signal IS sent for workflow object %s for transition %s -> %s" % (
                    self.workflow_object, self.transition_approval.source_state.label, self.transition_approval.destination_state.label))
        else:
            LOGGER.debug(
                "Post transition signal IS NOT sent for workflow object %s. Although transition_approval is occurred, not transition is occurred." % (
                    self.workflow_object))


class ProceedingSignal(object):
    def __init__(self, workflow_object, field_name, transition_approval):
        self.workflow_object = workflow_object
        self.field_name = field_name
        self.transition_approval = transition_approval

    def __enter__(self):
        pre_proceed.send(
            sender=ProceedingSignal.__class__,
            workflow_object=self.workflow_object,
            field_name=self.field_name,
            transition_approval=self.transition_approval,
        )
        LOGGER.debug("Pre transition approval signal is sent for workflow object %s for transition %s -> %s" % (
            self.workflow_object, self.transition_approval.source_state.label, self.transition_approval.destination_state.label))

    def __exit__(self, type, value, traceback):
        post_proceed.send(
            sender=TransitionSignal.__class__,
            workflow_object=self.workflow_object,
            field_name=self.field_name,
            transition_approval=self.transition_approval,
        )
        LOGGER.debug("Post transition approval signal is sent for workflow object %s for transition %s -> %s" % (
            self.workflow_object, self.transition_approval.source_state.label, self.transition_approval.destination_state.label))


class FinalSignal(object):
    def __init__(self, workflow_object, field_name):
        self.workflow_object = workflow_object
        self.field_name = field_name
        self.workflow = getattr(self.workflow_object.river, self.field_name)
        self.status = self.workflow.on_final_state

    def __enter__(self):
        if self.status:
            pre_final.send(
                sender=FinalSignal.__class__,
                workflow_object=self.workflow_object,
                field_name=self.field_name,
            )
            LOGGER.debug("Pre final signal is sent for workflow object %s for final state %s" % (self.workflow_object, self.workflow.get_state().label))

    def __exit__(self, type, value, traceback):
        if self.status:
            post_final.send(
                sender=TransitionSignal.__class__,
                workflow_object=self.workflow_object,
                field_name=self.field_name,
            )
            LOGGER.debug("Post final signal is sent for workflow object %s for final state %s" % (self.workflow_object, self.workflow.get_state().label))



import logging

from river.hooking.backends.base import powerset
from river.hooking.backends.memory import MemoryHookingBackend
from river.models.callback import Callback

__author__ = 'ahmetdal'

LOGGER = logging.getLogger(__name__)


class DatabaseHookingBackend(MemoryHookingBackend):

    def initialize_callbacks(self):
        self.__register(Callback.objects.filter(enabled=True))

    def __register(self, callback_objs):
        callbacks = []
        if callback_objs.exists():
            for callback in callback_objs:
                if callback.hash not in self.callbacks:
                    module, method_name = callback.method.rsplit('.', 1)
                    try:
                        method = getattr(__import__(module, fromlist=[method_name]), method_name, None)
                        if method:
                            self.callbacks[callback.hash] = method
                            callbacks.append(method)
                            LOGGER.debug("Callback '%s' from database is registered initially from database as method '%s' and module '%s'. " % (callback.hash, method_name, module))
                        else:
                            LOGGER.warning("Callback '%s' from database can not be registered. Because method '%s' is not in module '%s'. " % (callback.hash, method_name, module))
                    except ImportError:
                        LOGGER.warning("Callback '%s' from database can not be registered. Because module '%s'  does not exists. " % (callback.hash, module))
        return callbacks

    def register(self, hooking_cls, callback, workflow_object, field_name, override=False, *args, **kwargs):
        hash = super(DatabaseHookingBackend, self).register(hooking_cls, callback, workflow_object, field_name, override=override, *args, **kwargs)
        callback_obj, created = Callback.objects.update_or_create(
            hash=hash,
            defaults={
                'method': '%s.%s' % (callback.__module__, callback.__name__),
                'hooking_cls': '%s.%s' % (hooking_cls.__module__, hooking_cls.__name__),
            }
        )
        if created:
            LOGGER.debug("Callback '%s' is registered in database as method %s and module %s. " % (callback_obj.hash, callback.__name__, callback.__module__))
        else:
            LOGGER.debug("Callback '%s' is already registered in database as method %s and module %s. " % (callback_obj.hash, callback.__name__, callback.__module__))

        return hash

    def get_callbacks(self, hooking_cls, workflow_object, field_name, *args, **kwargs):
        callbacks = super(DatabaseHookingBackend, self).get_callbacks(hooking_cls, workflow_object, field_name, *args, **kwargs)
        if not callbacks:
            hashes = []
            for c in powerset(kwargs.keys()):
                skwargs = {}
                for f in c:
                    skwargs[f] = kwargs.get(f)
                hash = self.get_hooking_class(hooking_cls).get_hash(workflow_object, field_name, **skwargs)
                hashes.append(self.get_hooking_class_prefix(self.get_hooking_class(hooking_cls)) + hash)
            callbacks = self.__register(Callback.objects.filter(hash__in=hashes))
        return callbacks



from river.hooking.hooking import Hooking
from river.signals import pre_final, post_final

__author__ = 'ahmetdal'


class PreCompletedHooking(Hooking):
    pass


class PostCompletedHooking(Hooking):
    pass


pre_final.connect(PreCompletedHooking.dispatch)
post_final.connect(PostCompletedHooking.dispatch)



from __future__ import unicode_literals

from django.db import models
from django.db.models import CASCADE
from django.db.models.signals import post_save
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

from river.config import app_config
from river.models import State
from river.models.base_model import BaseModel
from river.models.managers.transitionmetada import TransitionApprovalMetadataManager

__author__ = 'ahmetdal'

BACKWARD = 0
FORWARD = 1

DIRECTIONS = [
    (BACKWARD, _("Backward")),
    (FORWARD, _("Forward"))
]


@python_2_unicode_compatible
class TransitionApprovalMeta(BaseModel):
    class Meta:
        app_label = 'river'
        verbose_name = _("Transition Approval Meta")
        verbose_name_plural = _("Transition Approval Meta")
        unique_together = [('content_type', 'field_name', 'source_state', 'destination_state', 'priority')]

    objects = TransitionApprovalMetadataManager()

    content_type = models.ForeignKey(app_config.CONTENT_TYPE_CLASS, verbose_name=_('Content Type'), on_delete=CASCADE)
    field_name = models.CharField(_("Field Name"), max_length=200)

    # transition = models.ForeignKey(Transition, verbose_name=_('Transition'), on_delete=CASCADE)
    source_state = models.ForeignKey(State, verbose_name=_("Source State"), related_name='transition_approval_meta_as_source', on_delete=CASCADE)
    destination_state = models.ForeignKey(State, verbose_name=_("Next State"), related_name='transition_approval_meta_as_destination', on_delete=CASCADE)
    # direction = models.SmallIntegerField(_("Transition Direction"), choices=DIRECTIONS, default=FORWARD)

    permissions = models.ManyToManyField(app_config.PERMISSION_CLASS, verbose_name=_('Permissions'), blank=True)
    groups = models.ManyToManyField(app_config.GROUP_CLASS, verbose_name=_('Groups'), blank=True)
    priority = models.IntegerField(default=0, verbose_name=_('Priority'), null=True)
    action_text = models.TextField(_("Action Text"), max_length=200, null=True, blank=True)
    parents = models.ManyToManyField('self', verbose_name='parents', related_name='children', symmetrical=False, db_index=True, blank=True)

    def natural_key(self):
        return self.field_name, self.content_type, self.source_state, self.destination_state, self.priority

    def __str__(self):
        return 'Field Name:%s, %s -> %s,Permissions:%s, Groups:%s, Order:%s' % (
            self.field_name,
            self.source_state,
            self.destination_state,
            ','.join(self.permissions.values_list('name', flat=True)),
            ','.join(self.groups.values_list('name', flat=True)), self.priority)


#
#
# def post_group_change(sender, instance, *args, **kwargs):
#     from river.services.proceeding import ProceedingService
#     from river.models.proceeding import PENDING
#
#     for proceeding_pending in instance.proceedings.filter(status=PENDING):
#         ProceedingService.override_groups(proceeding_pending, instance.groups.all())
#
#
# def post_permissions_change(sender, instance, *args, **kwargs):
#     from river.models.proceeding import PENDING
#     from river.services.proceeding import ProceedingService
#
#     for proceeding_pending in instance.proceedings.filter(status=PENDING):
#         ProceedingService.override_permissions(proceeding_pending, instance.permissions.all())


def post_save_model(sender, instance, *args, **kwargs):
    parents = TransitionApprovalMeta.objects \
        .filter(field_name=instance.field_name, destination_state=instance.source_state) \
        .exclude(pk__in=instance.parents.values_list('pk', flat=True)) \
        .exclude(pk=instance.pk)

    children = TransitionApprovalMeta.objects \
        .filter(field_name=instance.field_name, source_state=instance.destination_state) \
        .exclude(parents__in=[instance.pk]) \
        .exclude(pk=instance.pk)

    instance = TransitionApprovalMeta.objects.get(pk=instance.pk)
    if parents:
        instance.parents.add(*parents)

    for child in children:
        child.parents.add(instance)


# m2m_changed.connect(post_group_change, sender=TransitionApprovalMeta.groups.through)
# m2m_changed.connect(post_permissions_change, sender=TransitionApprovalMeta.permissions.through)

post_save.connect(post_save_model, sender=TransitionApprovalMeta)



import logging
from river.hooking.backends.base import BaseHookingBackend, powerset

__author__ = 'ahmetdal'

LOGGER = logging.getLogger(__name__)


class MemoryHookingBackend(BaseHookingBackend):
    def __init__(self):
        self.callbacks = {}

    def register(self, hooking_cls, callback, workflow_object, field_name, override=False, *args, **kwargs):
        hash = self.get_hooking_class_prefix(hooking_cls) + hooking_cls.get_hash(workflow_object, field_name, *args, **kwargs)
        if override or hash not in self.callbacks:
            self.callbacks[hash] = callback
            LOGGER.debug("Callback '%s'  is registered in memory as method '%s' and module '%s'. " % (hash, callback.__name__, callback.__module__))
        return hash

    def get_callbacks(self, hooking_cls, workflow_object, field_name, *args, **kwargs):
        callabacks = []
        for c in powerset(kwargs.keys()):
            skwargs = {}
            for f in c:
                skwargs[f] = kwargs.get(f)
            hash = self.get_hooking_class(hooking_cls).get_hash(workflow_object, field_name, **skwargs)
            callback = self.callbacks.get(self.get_hooking_class_prefix(self.get_hooking_class(hooking_cls)) + hash)
            if callback:
                callabacks.append(callback)
        return callabacks



from river.hooking.hooking import Hooking
from river.signals import pre_transition, post_transition

__author__ = 'ahmetdal'


class TransitionHooking(Hooking):

    @staticmethod
    def get_result_exclusions():
        return ["source_state", "destination_state"]

    @classmethod
    def get_hash(cls, workflow_object, field_name, source_state=None, destination_state=None, transition_approval=None, *args, **kwargs):
        return super(TransitionHooking, cls).get_hash(workflow_object, field_name) + ('source_state' + str(source_state.pk) if source_state else '') + (
            'destination_state' + str(destination_state.pk) if destination_state else '') + ('transition_approval' + str(transition_approval.pk) if transition_approval else '')


class PreTransitionHooking(TransitionHooking):
    pass


class PostTransitionHooking(TransitionHooking):
    pass


pre_transition.connect(PreTransitionHooking.dispatch)
post_transition.connect(PostTransitionHooking.dispatch)



import os
import sys

from django.test.utils import override_settings

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from django.core.management import call_command
from django.test import TestCase

_author_ = 'ahmetdal'


class TestMissingMigrations(TestCase):
    """
    This is the case to detect missing migration issues like https://github.com/javrasya/django-river/issues/30
    """

    migrations_before = []
    migrations_after = []

    def tearDown(self):
        """
            Remove migration file generated by test if there is any missing.
        """

        diff = list(set(self.migrations_after) - set(self.migrations_before))
        for d in diff:
            os.remove(os.path.join('river/migrations', d))

    @override_settings(MIGRATION_MODULES={})
    def test_missing_migrations(self):
        self.migrations_before = list(filter(lambda f: f.endswith('.py') and f != '__init__.py', os.listdir('river/migrations/')))

        out = StringIO()
        sys.stout = out

        call_command('makemigrations', 'river', stdout=out)

        self.migrations_after = list(filter(lambda f: f.endswith('.py') and f != '__init__.py', os.listdir('river/migrations/')))

        self.assertEqual("No changes detected in app 'river'\n", out.getvalue())

        self.assertEqual(len(self.migrations_before), len(self.migrations_after))



from django.db import models

__author__ = 'ahmetdal'


class TransitionApprovalMetadataManager(models.Manager):
    def get_by_natural_key(self, field_name, content_type, source_state, destination_state, priority):
        return self.get(field_name=field_name, content_type=content_type, source_state=source_state, destination_state=destination_state, priority=priority)



__author__ = 'ahmetdal'


class ErrorCode(object):
    NO_AVAILABLE_INITIAL_STATE = 1
    NO_AVAILABLE_FINAL_STATE = 2
    MULTIPLE_INITIAL_STATE = 3
    NO_AVAILABLE_NEXT_STATE_FOR_USER = 4
    INVALID_NEXT_STATE_FOR_USER = 5
    NEXT_STATE_IS_REQUIRED = 6

    MULTIPLE_STATE_FIELDS = 7
    NO_STATE_FIELD = 8



from itertools import combinations, chain

__author__ = 'ahmetdal'


def powerset(iterable):
    xs = list(iterable)
    # note we return an iterator rather than a list
    return chain.from_iterable(combinations(xs, n) for n in range(len(xs) + 1))


class BaseHookingBackend(object):
    def register(self, hooking_cls, callback, workflow_object, field_name, override=False, *args, **kwargs):
        raise NotImplementedError()

    def get_callbacks(self, hooking_cls, workflow_object, field_name, *args, **kwargs):
        raise NotImplementedError()

    @staticmethod
    def get_hooking_class(hooking_cls):
        if isinstance(hooking_cls, str):
            module, cls = hooking_cls.rsplit('.', 1)
            hooking_cls = getattr(__import__(module, fromlist=[cls]), cls)
        return hooking_cls

    @staticmethod
    def get_hooking_class_prefix(hooking_cls):
        return '%s.%s_' % (hooking_cls.__module__, hooking_cls.__name__)



import logging
import operator
from functools import reduce

from django.apps import AppConfig
from django.db.utils import OperationalError, ProgrammingError

__author__ = 'ahmetdal'

LOGGER = logging.getLogger(__name__)


class RiverApp(AppConfig):
    name = 'river'
    label = 'river'

    def ready(self):

        from river.hooking.backends.database import DatabaseHookingBackend
        from river.hooking.backends.loader import callback_backend

        for field_name in self._get_all_workflow_fields():
            try:
                transition_approval_meta = self.get_model('TransitionApprovalMeta').objects.filter(field_name=field_name)
                if transition_approval_meta.count() == 0:
                    LOGGER.warning("%s field doesn't seem have any transition approval meta in database. You should create it's TransitionApprovalMeta" % field_name)
            except (OperationalError, ProgrammingError):
                pass

        if isinstance(callback_backend, DatabaseHookingBackend):
            try:
                self.get_model('Callback').objects.exists()
                callback_backend.initialize_callbacks()
            except (OperationalError, ProgrammingError):
                pass
        LOGGER.debug('RiverApp is loaded.')

    @classmethod
    def _get_all_workflow_fields(cls):
        from river.core.workflowregistry import workflow_registry
        return reduce(operator.concat, map(list, workflow_registry.workflows.values()))



import os
import sys

__author__ = 'ahmetdal'

from setuptools import setup, find_packages

readme_file = os.path.join(os.path.dirname(__file__), 'README.rst')
try:
    long_description = open(readme_file).read()
except IOError as err:
    sys.stderr.write("[ERROR] Cannot find file specified as "
                     "``long_description`` (%s)\n" % readme_file)
    sys.exit(1)

setup(
    name='django-river',
    version='1.0.1',
    author='Ahmet DAL',
    author_email='ceahmetdal@gmail.com',
    packages=find_packages(),
    url='https://github.com/javrasya/django-river.git',
    description='Django Workflow Library',
    long_description=long_description,
    install_requires=[
        "Django",
        "mock",
        "factory-boy",
        "django-mptt"
    ],
    include_package_data=True,
    zip_safe=False,
    license='GPLv3',
    platforms=['any'],
)



import factory
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from factory import DjangoModelFactory

from river.models.state import State
from river.models.transitionapprovalmeta import TransitionApprovalMeta

__author__ = 'ahmetdal'


class ContentTypeObjectFactory(DjangoModelFactory):
    class Meta:
        model = ContentType

    model = factory.Sequence(lambda n: 'ect_model_%s' % n)


class UserObjectFactory(factory.DjangoModelFactory):
    class Meta:
        model = get_user_model()

    username = factory.Sequence(lambda n: 'User_%s' % n)

    @factory.post_generation
    def user_permissions(self, create, extracted, **kwargs):
        if not create:
            # Simple build, do nothing.
            return

        if extracted:
            # A list of groups were passed in, use them
            for permission in extracted:
                self.user_permissions.add(permission)

    @factory.post_generation
    def groups(self, create, extracted, **kwargs):
        if not create:
            # Simple build, do nothing.
            return

        if extracted:
            # A list of groups were passed in, use them
            for group in extracted:
                self.groups.add(group)


class GroupObjectFactory(factory.DjangoModelFactory):
    class Meta:
        model = 'auth.Group'

    name = factory.Sequence(lambda n: 'Group_%s' % n)

    @factory.post_generation
    def permissions(self, create, extracted, **kwargs):
        if not create:
            # Simple build, do nothing.
            return

        if extracted:
            # A list of groups were passed in, use them
            for permission in extracted:
                self.permissions.add(permission)


class PermissionObjectFactory(factory.DjangoModelFactory):
    class Meta:
        model = 'auth.Permission'

    name = factory.Sequence(lambda n: 'Permission_%s' % n)
    codename = factory.Sequence(lambda n: 'Codename_%s' % n)
    content_type = factory.SubFactory(ContentTypeObjectFactory)


class StateObjectFactory(DjangoModelFactory):
    class Meta:
        model = State

    label = factory.Sequence(lambda n: 's%s' % n)
    description = factory.Sequence(lambda n: 'desc_%s' % n)

class TransitionApprovalMetaFactory(DjangoModelFactory):
    class Meta:
        model = TransitionApprovalMeta

    source_state = factory.SubFactory(StateObjectFactory)
    destination_state = factory.SubFactory(StateObjectFactory)
    content_type = factory.SubFactory(ContentTypeObjectFactory)

    @factory.post_generation
    def permissions(self, create, extracted, **kwargs):
        if not create:
            # Simple build, do nothing.
            return

        if extracted:
            # A list of groups were passed in, use them
            for permission in extracted:
                self.permissions.add(permission)



from river.config import app_config

__author__ = 'ahmetdal'


def load_callback_backend():
    handler_module, hooking_cls = app_config.HOOKING_BACKEND_CLASS.rsplit('.', 1)
    return getattr(__import__(handler_module, fromlist=[hooking_cls]), hooking_cls)(**app_config.HOOKING_BACKEND_CONFIG)


callback_backend = load_callback_backend()



from django.db import models
from django.utils.translation import ugettext_lazy as _

from river.models.base_model import BaseModel

__author__ = 'ahmetdal'


class Callback(BaseModel):
    class Meta:
        app_label = 'river'
        verbose_name = _("Callback")
        verbose_name_plural = _("Callbacks")

    hash = models.CharField(_('Hash'), max_length=200, unique=True)
    method = models.CharField(_('Callback Method'), max_length=200)
    hooking_cls = models.CharField(_('HookingClass'), max_length=200)
    enabled = models.BooleanField(_('Enabled'), default=True)



class WorkflowRegistry(object):
    def __init__(self):
        self.workflows = {}
        self.class_index = {}

    def add(self, name, cls):
        self.workflows[id(cls)] = self.workflows.get(id(cls), set())
        self.workflows[id(cls)].add(name)
        self.class_index[id(cls)] = cls


workflow_registry = WorkflowRegistry()